/* proxyLL.c is part of proxyLL package.
   Copyright (C) 2009	KHI (bluedragon000@operamail.com)
  
*/

#include "includes.h"
#include "utils.h"
#include "tgc.h"
#include "session.h"

int initial_uid = 0;
int initial_gid = 0;

int timeout = 250;

const char *program_name;

// We try to be a nodaemon by default 
int	is_daemon = 0;

TGC	tgc;

/*------------------------------------------------------------------------------
   Prints version info.
--------------------------------------------------------------------------------*/
void print_version(void)
{
	printf("Proxy Listen-Listen, V%s Copyright (C) 2009 KHI \n", VERSION);
}

/*------------------------------------------------------------------------------
   Prints usage info.
--------------------------------------------------------------------------------*/
void print_usage(int exit_code)
{
	print_version();
	printf("Usage: %s  options ... \n\n", PACKAGE);
	printf("Options are:\n");
/*	
	printf("\n");
	printf(" -p, --port number	    The port to listen on for incomming connection from a user(default:%d)\n", TGC_USER_PORT);
	printf(" -m, --mysql  ip	    Mysql server ip (default: %s)\n", TGC_MYSQL_IP);
	printf(" -u, --user number	    Connected max user numbers(default: %d)\n", TGC_USER_MAX);
	printf(" -d, --daemon               Do become daemon\n");
	printf(" -l, --log file 	    Write logs to file. (default:'%s')\n", TGC_DEFAULT_LOGFILE);
	printf(" -g, --level number 	    Log level detail (default:%d).\n", TGC_DEFAULT_LOGLEVEL);
	printf(" -h, --help		    Display this.\n");
	printf(" -v, --version		    Display version number.\n\n");
	printf("\n");
*/
	printf(" -p  number	    The port to listen on for incomming connection from a user(default:%d)\n", TGC_USER_PORT);
	printf(" -s		    SSL-connection mode\n");
	printf(" -m  ip	            Mysql server ip \n");
	printf(" -u  number	    Connected max user numbers(default: %d)\n", TGC_USER_MAX);
	printf(" -d                 Do become daemon\n");
	printf(" -l  filename 	    Write logs to file. (default:'%s')\n", TGC_DEFAULT_LOGFILE);
	printf(" -g  number 	    Log level detail (default:%d).\n", TGC_DEFAULT_LOGLEVEL);
	printf(" -h  		    Display help.\n");
	printf(" -v  		    Display version number.\n\n");

	exit(exit_code);
}


/*-----------------------------------------------------------------------------
   Shutdown the server
------------------------------------------------------------------------------*/
RETSIGTYPE shutdown_server(int sig)
{
	PRINT_LOG(1, "Shutting down (signal:%d)", sig);
	tgc_shutdown(&tgc);
	exit(0);
}


/*-----------------------------------------------------------------------------
   Initialize the server data structures and variables.
------------------------------------------------------------------------------*/
void init_server(void)
{
	big_endian();

	if (tgc_post_init(&tgc)<0) {
		PRINT_LOG(1, "Failed to initialize!");
		print_usage(2);
	}

	if (tgc.isAUTH) {
		if (!chk_mysql_connect(&tgc)) {
			exit(2);
		}
	}
	
	initial_uid = geteuid();
	initial_gid = getegid();

	signal(SIGPIPE, SIG_IGN);

	signal(SIGTERM, shutdown_server);
	signal(SIGINT, shutdown_server);
	signal(SIGQUIT, shutdown_server);
	signal(SIGHUP, shutdown_server);
	signal(SIGCLD, sig_cld);
}


/*-----------------------------------------------------------------------------
  The Main function !
------------------------------------------------------------------------------*/
int main(int argc,char *argv[])
{
	int	loglevel     = TGC_DEFAULT_LOGLEVEL;
	int	next_option;
	extern	char *optarg;	/* getopt */
	extern  int  optind;	/* getopt */
	char	logfilename[MAX_PATH+1] = {0};
	char	temp[MAX_PATH] = {0};
	int	ntemp=0;
	struct in_addr inp;
	
	/* short options */
	const char *short_options = "p:u:m:l:sdhv";
	
	/* long options 
	const struct option long_options[] = {
		{"port",	1, NULL, 'p'},
		{"user",	1, NULL, 'u'},
		{"log",		1, NULL, 'l'},
		{"level",	1, NULL, 'g'},
		{"daemon",	0, NULL, 'd'},
		{"help",	0, NULL, 'h'},
		{"version",	0, NULL, 'v'},
		{NULL, 		0, NULL, 0}
	};
	*/

	strncpy(logfilename, TGC_DEFAULT_LOGFILE, MAX_PATH);
	tgc_pre_init(&tgc);

	program_name = argv[0];

	do {
		//next_option = getopt_long(argc, argv, short_options, long_options, NULL);
		next_option = getopt(argc, argv, short_options);

		switch (next_option) {
			case 'p':
				ntemp = atoi(optarg);
				if (!strnum(optarg) || ntemp<=0 || ntemp>=65535) {
					fprintf(stderr, "Invalid port number '%s'\n", optarg);
					exit(2);
				}
				tgc.port = ntemp;
				break;
			case 's':
				tgc.isSSL = 1;
				break;
			case 'u':
				ntemp = atoi(optarg);
				if (!strnum(optarg) || ntemp<=0 || ntemp>=127) {
					fprintf(stderr, "Invalid user numbers '%s'\n", optarg);
					exit(2);
				}
				tgc.user_max = ntemp;
				break;
			case 'l':
				strncpy(logfilename, optarg, MAX_PATH);
				break;

			case 'm':
				if (!inet_aton(optarg, &inp)) 
				{
					fprintf(stderr, "Invalid mysql ip '%s'\n", optarg);
					exit(2);
				}
				strcpy(tgc.mysql_ip, optarg);
				tgc.isAUTH = 1;
				break;
/*
			case 'g':
				ntemp = atoi(optarg);
				if (!strnum(optarg) || ntemp<0) {
					fprintf(stderr, "Invalid loglevel '%s'\n", optarg);
					exit(2);
				}
				loglevel = ntemp;
				break;
*/
			case 'd':
				is_daemon = 1;
				break;
			case 'h':
				print_usage(0);
				break;
			case 'v':
				print_version();
				exit(0);
				break;
			case -1:
				/* done with options! */
				break;
			case '?':
			default:
				print_usage(1);
		}
	} while (next_option != -1);
	
	/* if there are non-option parameters, print usage and exit! */

//	if (optind<argc || argc<=1) 
	if (argc<=1) 
		print_usage(1);

	init_log(loglevel, logfilename, is_daemon);

	init_server();

#ifdef HAVE_FORK
	if (is_daemon)
		become_daemon();
#endif
	tgc_run(&tgc);

	close_log();
	return 0;
}

