

/* tgc.c is part of proxyLL package.
   Copyright (C) 2009	KHI (bluedragon000@operamail.com)

*/

//#include <sys/mman.h>
#include "tgc.h"
#include "includes.h"
#include "utils.h"
#include "dbmysql.h"

static int s_server_session_id_context = 1;


/* Print SSL errors and exit*/
int berr_exit(char *string)
{
    PRINT_LOG(1, "%s\n", string);
    exit(0);
}

/*The password code is not thread safe*/
static int password_cb(char *buf, int num, int rwflag, void *userdata)
{
//    if(num<strlen(pass)+1)
    if(num<strlen(PASSWORD)+1)
      return(0);

    strcpy(buf,PASSWORD);
    return(strlen(PASSWORD));
}


/*-----------------------------------------------------------------------------
 Do the SSL context initialization
 ----------------------------------------------------------------------------*/
SSL_CTX *initialize_ctx(char *password)
{
    SSL_METHOD *meth;
    SSL_CTX *ctx;
	int h;

      /* Global system initialization*/
      SSL_library_init();
      SSLeay_add_ssl_algorithms();
      SSL_load_error_strings();
    /* Create our context*/
    meth=SSLv23_server_method();//SSLv23_method();
    ctx=SSL_CTX_new(meth);
    PRINT_LOG(1, "initialize SSL context");

    /* Load our keys and certificates*/
	if ((h=open(CERTF_ABS, O_RDONLY, S_IREAD)) == -1) {
		if(!(SSL_CTX_use_certificate_chain_file(ctx, CERTF)))
		berr_exit("Can't read certificate file!");
	}
	else {
		close(h);
		if(!(SSL_CTX_use_certificate_chain_file(ctx, CERTF_ABS)))
		berr_exit("Can't read certificate file.");
	}

    SSL_CTX_set_default_passwd_cb(ctx, password_cb);

	if ((h=open(KEYF_ABS, O_RDONLY, S_IREAD)) == -1) {
		if(!(SSL_CTX_use_PrivateKey_file(ctx, KEYF, SSL_FILETYPE_PEM)))
		berr_exit("Can't read key file");
	}
	else {
		close(h);
		if(!(SSL_CTX_use_PrivateKey_file(ctx, KEYF_ABS, SSL_FILETYPE_PEM)))
		berr_exit("Can't read key file");
	}

    if (!SSL_CTX_check_private_key(ctx)) 
      berr_exit("ssl certificate & privatekey file not matching!");

    /* Load the CAs we trust*/
//     if(!(SSL_CTX_load_verify_locations(ctx, CA_LIST,0)))
//       berr_exit("Can't read CA list");
// #if (OPENSSL_VERSION_NUMBER < 0x00905100L)
//     SSL_CTX_set_verify_depth(ctx,1);
// #endif

    return ctx;
}

/*-----------------------------------------------------------------------------
 Do the 1st phase initialization
 return value (0: okay, else: not okay)
-----------------------------------------------------------------------------*/
int tgc_pre_init(TGC *tgc)
{
	if (!tgc) {
		PRINT_LOG(1, "Internal Error!");
		exit(1);
		//return E_TGC_IE;
	}

	memset(tgc, 0, sizeof(TGC));
	
	return 0;
}



/*-----------------------------------------------------------------------------
 Do the 2nd phase initialization
 return value (0: okay, else: not okay)
-----------------------------------------------------------------------------*/
int tgc_post_init(TGC *tgc)
{
	int rc=0;

	if (!tgc) {
		PRINT_LOG(1, "Internal Error!");
		exit(1);
		//return E_TGC_IE;
	}

	if ( tgc->port == 0 )
		tgc->port = TGC_USER_PORT;

	if (!tgc->user_max)
		tgc->user_max = TGC_USER_MAX;

	if (tgc->isSSL)	{
		/* SSL preliminaries. We keep the certificate and key with the context. */
		tgc->ctx = initialize_ctx(PASSWORD);
		if (!tgc->ctx) {
			berr_exit("ssl context create Error!");
		}
	}

	return rc;
}

/*-----------------------------------------------------------------------------
 Ask CC for a new connection
-----------------------------------------------------------------------------*/
int tgc_send(int sd, char *cmd)
{
	if (write(sd, cmd, strlen(cmd)) < 1) {
		PRINT_LOG(3, "Failed sending connection request");
		return E_TGC_NOCANDO;
	}
	return 0;
}
/////////////////////////////////////////
int tgc_write(socket_queue *sq, char *cmd, int len, int isSSL)
{
	int 	err;
	if (isSSL) 
		err = SSL_write(sq->ssl, cmd, len);
	else
		err = write(sq->sd, cmd, len);

	if (err < 1) {
		PRINT_LOG(3, "Failed sending connection request");
		return E_TGC_NOCANDO;
	}

	return err;
}
/*-----------------------------------------------------------------------------
 Read internal command!
-----------------------------------------------------------------------------*/
int tgc_read(socket_queue *sq, char *buf, int isSSL)
{
	int read_size, val;

	if (isSSL) 
		read_size = SSL_read(sq->ssl, buf, TGC_BUFFER_SIZE);
	else
		read_size = read(sq->sd, buf, TGC_BUFFER_SIZE);

	if (read_size == 0) 
		return E_TGC_NOCANDO;
	else if (read_size < 0) {
		PRINT_LOG(3, "Failed reading command  ");
		return E_TGC_NOCANDO;
	}

	buf[read_size] = '\0';
	return read_size;
}

/*-----------------------------------------------------------------------------
 * Add an element to queue
-----------------------------------------------------------------------------*/
int tgc_add_queue(socket_queue **list, int sd, SSL* argssl, unsigned long sip)
{
	socket_queue	*tq;

	if (sd<0)
		return E_TGC_NOCANDO;
	
	if ( !(tq=(socket_queue *) malloc(sizeof(socket_queue))) ) {
		PRINT_LOG(1, "Out of memory!");
		return E_TGC_IE;
	}

	tq->sd = sd;
	tq->ssl = argssl;
	tq->ip = sip;
	tq->si = get_current();
	tq->isUser = 0;
	tq->sn = 0;
	tq->retry = 0;
	tq->userlen = 0;
	memset(tq->username, 0, MAX_USERNAME);
	tq->prev = NULL;

	if (*list) {
		tq->next = *list;
		(*list)->prev = tq;
		*list = tq;
	} else {
		tq->next = NULL;
		*list = tq;
	}
	return 0;
}


/*-----------------------------------------------------------------------------
 Remove an element from queue
-----------------------------------------------------------------------------*/
int  tgc_remove_queue(socket_queue **list, socket_queue *sq, SSL **ssl)
{
	int	sd;

	if (!sq || !*list)
		return E_TGC_NOCANDO;

	sd = sq->sd;
	*ssl = sq->ssl;

	if (sq->prev) {
		if (sq->next) { // middle node
			sq->prev->next = sq->next;
			sq->next->prev = sq->prev;
		} else { // end node
			sq->prev->next = NULL;
		}
	} else {
		if (sq->next) { // first node
			*list = sq->next;
			(*list)->prev = NULL;
		} else { //the only node
			(*list) = NULL;
		}
	}
	free(sq);

	return(sd);
}

void my_ssl_shutdown(SSL *ssl, int sd)
{
    int r;

	if (ssl) {
		r=SSL_shutdown(ssl);
		if(!r){
		  /* If we called SSL_shutdown() first then we always get return value of '0'. In
			 this case, try again, but first send a TCP FIN to trigger the other side's 
			 close_notify*/
		  shutdown(sd,1);
		  r=SSL_shutdown(ssl);
		}

		switch(r){  
		  case 1:
			break; /* Success */
		  case 0:
		  case -1:
		  default:
			PRINT_LOG(3, "Shutdown failed");
		}
		SSL_free(ssl);
	}
	else {
		  shutdown(sd,1);
	}
	close(sd);
}

/*-----------------------------------------------------------------------------
 Shutdown, whatelse?!
-----------------------------------------------------------------------------*/
void tgc_shutdown(TGC *tgc)
{
	int	sd=-1;
	SSL 	*ssl;
	
	if (sd>=0) 
		close(sd);

	if (tgc->sd_accept>=0)
		close(tgc->sd_accept);

	while (tgc->socketq) {
		sd = tgc_remove_queue(&(tgc->socketq), tgc->socketq, &ssl);
		my_ssl_shutdown(ssl, sd);
		
	}
	SSL_CTX_free (tgc->ctx);
}


/*------------------------------------------------------------------------------
 *   Signal handler for SIGPIPE (write on a disconnected socket)
 *   --------------------------------------------------------------------------------*/
void do_abort(void )
{
        PRINT_LOG(1,"Aborted.");
        exit(1);
}
 

/*-----------------------------------------------------------------------------
 * Transfer data from one socket to another through the normal sorcket
------------------------------------------------------------------------------*/
int tgc_rxtx_from_socket(int rx_sd, int tx_sd)
{
	int	nread=0, nwritten=0, i;
	char	buf[TGC_BUFFER_SIZE+1];

	memset(buf, 0, TGC_BUFFER_SIZE);

	if (rx_sd<0 || tx_sd<0)
		return E_TGC_IE;

	nread = read(rx_sd, buf, TGC_BUFFER_SIZE);

	if (nread < 1 ) {
		if (nread < 0) PRINT_LOG(3, "Error reading from socket");
			return E_TGC_READ;
	}
	nwritten = write(tx_sd, buf, nread);

	if ( nwritten < nread  ) {
		PRINT_LOG(3, "Error writing to socket");
		return E_TGC_WRITE;
	}

	return 0;
}

/*-----------------------------------------------------------------------------
 * Transfer data from one socket to another through SSL
------------------------------------------------------------------------------*/
int tgc_rxtx_from_SSL (SSL *rx_ssl, SSL *tx_ssl)
{
	int	nread=0, nwritten=0, i;
	char	buf[TGC_BUFFER_SIZE+1];

	memset(buf, 0, TGC_BUFFER_SIZE);

	if ((!rx_ssl) || (!tx_ssl))
		return E_TGC_IE;

	nread = SSL_read(rx_ssl, buf, TGC_BUFFER_SIZE);

	switch(SSL_get_error(rx_ssl, nread)) {
		case SSL_ERROR_NONE:
				break;
		case SSL_ERROR_WANT_READ:
				PRINT_LOG(3, "read more");
				break;
		case SSL_ERROR_ZERO_RETURN:
				return E_TGC_READ;
		default:
				PRINT_LOG(3, "SSL read error");
				return E_TGC_READ;
				break;
	}

//PRINT_LOG(3, "read %d", nread);
//if (nread > 32000) PRINT_LOG(3, "size over %d",nread);
	nwritten = SSL_write(tx_ssl, buf, nread);

	switch(SSL_get_error(tx_ssl, nwritten)) {
		case SSL_ERROR_NONE:
				break;
		case SSL_ERROR_WANT_WRITE:
				PRINT_LOG(3, "write more");
				break;
		default:
				// End of data
				PRINT_LOG(3, "SSL write error!");
				return E_TGC_WRITE;
				break;
	}
//PRINT_LOG(3, "write %d", nwritten);

	return 0;
}


/*-----------------------------------------------------------------------------
 * Pumping from one socket to another (desk-owner)
------------------------------------------------------------------------------*/
int tgc_pump(int sdi, int sdx, SSL *ssli, SSL *sslx, unsigned long desk_ip, unsigned long owner_ip)
{
	fd_set	rfds, reads;
	int	rc=0;
	int 	*sock_count;
	struct	in_addr   print_ip, print_ip2;

	if (sdi<0 || sdx<0)
		return E_TGC_IE;

	sock_count = ipc_connect();

	FD_ZERO(&rfds);
	FD_SET(sdi, &rfds);
	FD_SET(sdx, &rfds);
	while (sdi>=0 && sdx>=0) {
		reads = rfds; 
		if ( select(FD_SETSIZE, &reads, NULL, NULL, NULL) < 0 ) 
			continue;

		if (FD_ISSET(sdi, &reads)) {
			print_ip.s_addr = desk_ip;
			if (ssli)  {
				rc=tgc_rxtx_from_SSL(ssli, sslx);
				//PRINT_LOG(3, "send from desk(%s)", inet_ntoa(print_ip));
				if (rc < 0 ) {
					PRINT_LOG(3, "disconnect from desk(%s)", inet_ntoa(print_ip));
					my_ssl_shutdown(ssli, sdi);
					my_ssl_shutdown(sslx, sdx);
					break;//return rc;
				}
			}
			else	{
				rc=tgc_rxtx_from_socket(sdi, sdx);
				if (rc < 0 ) {
					PRINT_LOG(3, "disconnect from desk(%s)", inet_ntoa(print_ip));
					close(sdi);
					close(sdx);
					break;//return rc;
				}
			}
		}

		if (FD_ISSET(sdx, &reads)) {
			print_ip.s_addr = owner_ip;
			if (sslx)	{
				rc=tgc_rxtx_from_SSL(sslx, ssli);
				//PRINT_LOG(3, "send from owner(%s)", inet_ntoa(print_ip));
				if (rc  < 0 ) {
					PRINT_LOG(3, "disconnect from owner(%s)", inet_ntoa(print_ip));
					my_ssl_shutdown(sslx, sdx);
					my_ssl_shutdown(ssli, sdi);
					break;//return rc;
				}
			}
			else	{
				rc=tgc_rxtx_from_socket(sdx, sdi);
				if (rc  < 0 ) {
					PRINT_LOG(3, "disconnect from owner(%s)", inet_ntoa(print_ip));
					close(sdx);
					close(sdi);
					break;//return rc;
				}
			}
		}
	}
	*sock_count -= 2;
	return rc;
}

void	tgc_close(socket_queue **socketq, socket_queue *sq, fd_set *rfds, int *count)
{
	int	sd;
	SSL 	*ssl;

	sd = tgc_remove_queue(socketq, sq, &ssl);
	my_ssl_shutdown(ssl, sd);

	FD_CLR(sd, rfds);
	--(*count);
	PRINT_LOG(3, "Connection close request !");
}


/*-----------------------------------------------------------------------------
 * Start the main scheme
------------------------------------------------------------------------------*/
int tgc_run(TGC *tgc) 
{
	SSL		*ssl, *sslo, *sslc;
	fd_set		rfds, reads;
	sigset_t	sigmask;
	struct		sockaddr_in addr;
	struct		in_addr   print_ip, print_ip2;
	socklen_t 	in_addrlen = sizeof(addr);
	int		sdo, sdc, sd, st, val;
	socket_queue	*ptrq, *tmpq, *opq, *prev_sq;
	long 		session_number, sn, hash;
	int		*sock_count, i;
	int		isFail, isFind;
	int		timeout;
	int		o_flag;
	long		now, delta;
	unsigned long	owner_ip, desk_ip;
	short		*shortptr;	

	if (!tgc)
		return E_TGC_IE;

	if ( (tgc->sd_accept=open_server_socket(tgc->port)) < 0 ) {
		PRINT_LOG(1, "Can't bind to port %d", tgc->port);
		return E_TGC_NOCANDO;
	}
	PRINT_LOG(3, "waiting for user on port %d ...", tgc->port);
	
	sock_count = ipc_connect();

	FD_ZERO(&rfds);
	FD_SET(tgc->sd_accept, &rfds);
	*sock_count = 0;

	while(1) 
	{
		reads = rfds;


		if ( select(FD_SETSIZE, &reads, NULL, NULL, NULL) < 0 ) 
			continue;
//printf("\n user max = %d , connected users = %d ", tgc->user_max, *sock_count);
		// incoming client connection
		if (FD_ISSET(tgc->sd_accept, &reads)) { 
			sd = accept_connection(tgc->sd_accept, (struct sockaddr_in *)&addr, (socklen_t *)&in_addrlen);
			if (sd==-1) {
				if (errno==EINTR)
					continue;
				PRINT_LOG(1, "Error accepting new client connection");
				break; // exit
			}
#ifdef HAVE_INET_NTOA
			PRINT_LOG(3, " Connected client from %s", inet_ntoa(addr.sin_addr));
#else
			PRINT_LOG(3, "Client connected");
#endif
			/* tell client to what is socket type */
			if (tgc->isSSL) 
				tgc_send(sd, SOCKET_SSL);
			else
				tgc_send(sd, SOCKET_NORMAL);
			/* tell client to whether authintication mode or not */
			if (tgc->isAUTH) 
				tgc_send(sd, STATUS_AUTH);
			else
				tgc_send(sd, STATUS_NONAUTH);

			/* TCP connection is ready. Do server side SSL.    */
			isFail = 0;
			ssl = NULL;
			if (tgc->isSSL) {
				ssl = SSL_new (tgc->ctx);
				if (!ssl) {
					PRINT_LOG(1, "Error SSL create");
					return E_TGC_NOCANDO; 
				}
				/* Use a non blocking mode for protect of delay by non-ssl connection */
				o_flag = fcntl(sd, F_GETFL);
				fcntl(sd, F_SETFL, o_flag | O_NONBLOCK);
				SSL_set_fd (ssl, sd);
				now = get_current();
				timeout = 0;
				do {
					delta = get_current() - now;
					if (SSL_accept(ssl) != -1) {
						timeout = 1; 
						break;
					}
				}
				while (delta < TGC_TIMEOUT);
				fcntl(sd, F_SETFL, o_flag);
				if (!timeout) {
					PRINT_LOG(1, "Error ssl connect timeout");
					isFail = 1;
				}
			}

			/* Check if users exceed the max limit */
			if (*sock_count > tgc->user_max) {
				if (tgc->isSSL)
					SSL_write(ssl, STATUS_FULL, strlen(STATUS_FULL));
				else
					tgc_send(sd, STATUS_FULL);
				PRINT_LOG(3, "Full out connections");
				isFail = 1;
			}
			
			if (isFail) {
				close(sd);
				if (ssl) SSL_free(ssl);
			}
			else {
				tgc_add_queue( &(tgc->socketq), sd, ssl, addr.sin_addr.s_addr);
//printf("\n connected user = %x ", tgc->socketq->ip);
				FD_SET(sd, &rfds);
				++(*sock_count);
			}
			continue;
		}	
		// response with userID(from owner) or session number(from desk)
		ptrq = tgc->socketq;
		while (ptrq)
		{
		// check timeout from connect
		/*	
		    if (ptrq->isUser == 0) {
			now = get_current();
			delta = now - ptrq->si;
			if (delta >  TGC_TIMEOUT*20) {
				print_ip.s_addr = ptrq->ip;
				PRINT_LOG(1, "connect timeout. it could be not a normal user(%s)",inet_ntoa(print_ip));
				prev_sq = ptrq;
				ptrq = ptrq->next;
				sd = tgc_remove_queue(&(tgc->socketq), prev_sq, &ssl);
				FD_CLR(sd, &rfds);
				my_ssl_shutdown(ssl, sd);
				--(*sock_count);
				continue;
			}
		    }
		*/
		    if (FD_ISSET(ptrq->sd, &reads)) {
			
			memset(tgc->buf, '\0', TGC_BUFFER_SIZE);
			val = tgc_read(ptrq, tgc->buf, tgc->isSSL);

			// check the abnormal traffic
			if ((val == E_TGC_NOCANDO) || (tgc->buf[0] != IS_OWNER && tgc->buf[0] != IS_DESK))  {
				print_ip.s_addr = ptrq->ip;
				if (val == E_TGC_NOCANDO)
					PRINT_LOG(3, "disconnected client from %s", inet_ntoa(print_ip));
				else
					PRINT_LOG(3, "non remote client from user(%s)", inet_ntoa(print_ip));
				prev_sq = ptrq;
				ptrq = ptrq->next;
				sd = tgc_remove_queue(&(tgc->socketq), prev_sq, &ssl);
				FD_CLR(sd, &rfds);
				my_ssl_shutdown(ssl, sd);
				--(*sock_count);
				continue;
			}

			// check owner
			if ((tgc->buf[0] == IS_OWNER) && (ptrq->sn == 0)) {
				// detect delimiter between username and password
				shortptr = (short *) tgc->buf;
				for (i=1, isFind=0; i < val/2; i++)
				{
//PRINT_LOG(1, "recv <%x> (%d)", shortptr[i], i);
					if (shortptr[i] == 0x0A) {
						ptrq->userlen = i * 2;
					}
				}
PRINT_LOG(1, "recv userlen (%d)", ptrq->userlen);

				if (!ptrq->userlen) {
					print_ip.s_addr = ptrq->ip;
					PRINT_LOG(3, " bad username from owner(%s)", inet_ntoa(print_ip));
					goto error_auth;
				}

				memcpy(ptrq->username, &tgc->buf[2], ptrq->userlen);
				ptrq->isUser = 1;
				//hash = generate_hash(tgc->buf, val)) != 0)	// match
				if (tgc->isAUTH) {
					if ((session_number = query_username(tgc, val)) != 0)	// match
					{
						print_ip.s_addr = ptrq->ip;
						PRINT_LOG(3, " login owner from %s", inet_ntoa(print_ip));
						ptrq->sn = session_number;
						sprintf(tgc->buf, "O%u", session_number);
						tgc_write(ptrq, tgc->buf, 9, tgc->isSSL);
					}
					else 
					{
						print_ip.s_addr = ptrq->ip;
						PRINT_LOG(3, " failed authenticate from owner(%s)", inet_ntoa(print_ip));
error_auth:
						ptrq->retry += 1;
						tgc_write(ptrq, STATUS_ERROR, 1, tgc->isSSL);
						if (ptrq->retry >= TGC_RETRY)
						{
							prev_sq = ptrq;
							ptrq = ptrq->next;
							sd = tgc_remove_queue(&(tgc->socketq), prev_sq, &ssl);
							FD_CLR(sd, &rfds);
							my_ssl_shutdown(ssl, sd);
							--(*sock_count);
							continue;
						}
					}
				}
				else {
					ptrq->sn = generate_session_number2(8);
					memcpy(&tgc->buf[1], ptrq->username, ptrq->userlen);
					sprintf(tgc->buf, "O%u", ptrq->sn);
					tgc_write(ptrq, tgc->buf, 9, tgc->isSSL);
	
					print_ip.s_addr = ptrq->ip;
					PRINT_LOG(3, " login owner from %s (%u)", inet_ntoa(print_ip), ptrq->sn);
				}
			}

			// response from desk
			if (tgc->buf[0] == IS_DESK) {
				ptrq->isUser = 1;

				if (val < SESSION_DIGIT)	
					strcpy(tgc->buf, "C-9999999");

				session_number = atol(&(tgc->buf[1]));
				//PRINT_LOG(3, "desk sn = %d", session_number);

				// interprete and query session number
				tmpq = tgc->socketq;
				opq = NULL;
				if (session_number) {
					while (tmpq)
					{
						if (tmpq->sn == session_number)
						{
							opq = tmpq;
							print_ip.s_addr = ptrq->ip;
							PRINT_LOG(3, " login desk from %s (%d)", inet_ntoa(print_ip), session_number);
							break;
						}
						tmpq = tmpq->next;
					}
				}
				if (opq)
				{
					// notify peer-startup to desk and owner
					tgc_write(ptrq, STATUS_OK, 1, tgc->isSSL);
					tgc_write(ptrq, opq->username, opq->userlen, tgc->isSSL);
					tgc_write(opq, STATUS_OK, 1, tgc->isSSL);

					owner_ip = opq->ip;
					desk_ip = ptrq->ip;
					sdo = tgc_remove_queue(&(tgc->socketq), opq, &sslo);
					FD_CLR(sdo, &rfds);
					sdc = tgc_remove_queue(&(tgc->socketq), ptrq, &sslc);
					FD_CLR(sdc, &rfds);
					// send the Start command to owner and desk
	
					if ((sdc>=0) && (sdo>=0))
					{
						print_ip.s_addr = owner_ip;
						print_ip2.s_addr = desk_ip;
						PRINT_LOG(3, "peer-owner %s (session %d)", inet_ntoa(print_ip), session_number);
						PRINT_LOG(3, "peer-desk  %s (session %d)", inet_ntoa(print_ip2), session_number);
						// FORK
						if (fork()==0) 
						{ // child
							// close the other socket descriptors
							close(tgc->sd_accept);
							tmpq = tgc->socketq;
							while (tmpq)
							{
								if (tmpq->sd > 0) {
									close(tmpq->sd);
									if (tmpq->ssl) SSL_free (tmpq->ssl);
								}
								tmpq = tmpq->next;
							}
							// start the communication between owner and desk
							tgc_pump(sdc, sdo, sslc, sslo, desk_ip, owner_ip);
							exit(0);//break;
						}
						else 
						{ // parent
							close(sdc);
							close(sdo);
							if (sslc) SSL_free (sslc);
							if (sslo) SSL_free (sslo);
						}
					}
					//ptrq = tgc->socketq;
					break;
				}
				else
				{
					PRINT_LOG(3, "peer connection FAIL");
					tgc_write(ptrq, STATUS_ERROR, 1, tgc->isSSL);
					ptrq->retry += 1;
					if (ptrq->retry >= TGC_RETRY)
					{
						print_ip.s_addr = ptrq->ip;
						PRINT_LOG(3, " rejected desk(%s)", inet_ntoa(print_ip));
						prev_sq = ptrq;
						ptrq = ptrq->next;
						sd = tgc_remove_queue(&(tgc->socketq), prev_sq, &ssl);
						FD_CLR(sd, &rfds);
						my_ssl_shutdown(ssl, sd);
						--(*sock_count);
						continue;
					}
				}
			}
		    }
		    ptrq = ptrq->next;
		}
	}
	return 0;
}






